# -*- coding: utf-8 -*-

project = 'toasty'
author = 'Chris Beaumont and the AAS WorldWide Telescope Team'
copyright = '2014-2020, ' + author

release = '0.dev0'  # cranko project-version
version = '.'.join(release.split('.')[:2])

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.smart_resolver',
    'numpydoc',
]

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
language = None
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
pygments_style = 'sphinx'
todo_include_todos = False

html_theme = 'bootstrap-astropy'
html_theme_options = {
    'logotext1': 'toasty',
    'logotext2': '',
    'logotext3': ':docs',
    'astropy_project_menubar': False,
}
html_static_path = ['_static']
htmlhelp_basename = 'toastydoc'

intersphinx_mapping = {
    'python': (
        'https://docs.python.org/3/',
        (None, 'http://data.astropy.org/intersphinx/python3.inv')
    ),

    'astropy': (
        'https://docs.astropy.org/en/stable/',
        None
    ),

    'numpy': (
        'https://numpy.org/doc/stable/',
        (None, 'http://data.astropy.org/intersphinx/numpy.inv')
    ),

    'PIL': (
        'https://pillow.readthedocs.io/en/stable/',
        None
    ),

    'wwt_data_formats': (
        'https://wwt-data-formats.readthedocs.io/en/stable/',
        None
    ),
}

numpydoc_show_class_members = False

nitpicky = True
nitpick_ignore = [
    ('py:class', 'ipywidgets.widgets.domwidget.DOMWidget')
]

default_role = 'obj'

html_logo = 'images/logo.png'

linkcheck_retries = 5
linkcheck_timeout = 10



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2021 the AAS WorldWide Telescope project.
# Licensed under the MIT License.

"""Entrypoint for the "toasty" command-line interface.

"""

from __future__ import absolute_import, division, print_function

__all__ = """
die
entrypoint
warn
""".split()

import argparse
import os.path
import sys
import warnings

from wwt_data_formats.cli import EnsureGlobsExpandedAction


# General CLI utilities


def die(msg):
    print("error:", msg, file=sys.stderr)
    sys.exit(1)


def warn(msg):
    print("warning:", msg, file=sys.stderr)


# "cascade" subcommand


def cascade_getparser(parser):
    parser.add_argument(
        "--parallelism",
        "-j",
        metavar="COUNT",
        type=int,
        help="The parallelization level (default: use all CPUs; specify `1` to force serial processing)",
    )
    parser.add_argument(
        "--format",
        "-f",
        metavar="FORMAT",
        default=None,
        choices=["png", "jpg", "npy", "fits"],
        help="The format of data files to cascade. If not specified, this will be guessed.",
    )
    parser.add_argument(
        "--start",
        metavar="DEPTH",
        type=int,
        help="The depth of the TOAST layer to start the cascade",
    )
    parser.add_argument(
        "pyramid_dir",
        metavar="DIR",
        help="The directory containing the tile pyramid to cascade",
    )


def cascade_impl(settings):
    from .merge import averaging_merger, cascade_images
    from .pyramid import PyramidIO

    pio = PyramidIO(settings.pyramid_dir, default_format=settings.format)

    start = settings.start
    if start is None:
        die("currently, you must specify the start layer with the --start option")

    cascade_images(
        pio, start, averaging_merger, parallel=settings.parallelism, cli_progress=True
    )


# "check-avm" subcommand


def check_avm_getparser(parser):
    parser.add_argument(
        "--print",
        "-p",
        dest="print_avm",
        action="store_true",
        help="Print the AVM data if present",
    )
    parser.add_argument(
        "--exitcode",
        action="store_true",
        help="Exit with code 1 if no AVM tags are present",
    )
    parser.add_argument(
        "image",
        metavar="PATH",
        help="An image to check for AVM tags",
    )


def check_avm_impl(settings):
    try:
        from pyavm import AVM, exceptions

        SYNTAX_EXCEPTIONS = (
            exceptions.AVMListLengthError,
            exceptions.AVMItemNotInControlledVocabularyError,
            exceptions.AVMEmptyValueError,
        )
    except ImportError:
        die("cannot check AVM tags: you must install the `pyavm` package")

    try:
        # PyAVM can issue a lot of warnings that aren't helpful to the user.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            avm = AVM.from_image(settings.image)
    except exceptions.NoXMPPacketFound:
        print(f"{settings.image}: no AVM tags")
        sys.exit(1 if settings.exitcode else 0)
    except SYNTAX_EXCEPTIONS as e:
        print(
            f"{settings.image}: AVM data are present but malformed: {e} ({e.__class__.__name__})"
        )
        sys.exit(1)
    except IOError as e:
        die(f"error probing AVM tags of `{settings.image}`: {e}")

    if settings.print_avm:
        print(f"{settings.image}: AVM tags are present")
        print()
        print(avm)
    else:
        print(f"{settings.image}: AVM tags are present (use `--print` to display them)")


# "make_thumbnail" subcommand


def make_thumbnail_getparser(parser):
    from .image import ImageLoader

    ImageLoader.add_arguments(parser)

    parser.add_argument(
        "imgpath",
        metavar="IN-PATH",
        help="The image file to be thumbnailed",
    )
    parser.add_argument(
        "outpath",
        metavar="OUT-PATH",
        help="The location of the new thumbnail file",
    )


def make_thumbnail_impl(settings):
    from .image import ImageLoader

    olp = settings.outpath.lower()
    if not (olp.endswith(".jpg") or olp.endswith(".jpeg")):
        warn(
            'saving output in JPEG format even though filename is "{}"'.format(
                settings.outpath
            )
        )

    img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
    thumb = img.make_thumbnail_bitmap()

    with open(settings.outpath, "wb") as f:
        thumb.save(f, format="JPEG")


# "tile_allsky" subcommand


def tile_allsky_getparser(parser):
    from .image import ImageLoader

    ImageLoader.add_arguments(parser)

    parser.add_argument(
        "--name",
        metavar="NAME",
        default="Toasty",
        help="The image name to embed in the output WTML file (default: %(default)s)",
    )
    parser.add_argument(
        "--outdir",
        metavar="PATH",
        default=".",
        help="The root directory of the output tile pyramid (default: %(default)s)",
    )
    parser.add_argument(
        "--placeholder-thumbnail",
        action="store_true",
        help="Do not attempt to thumbnail the input image -- saves memory for large inputs",
    )
    parser.add_argument(
        "--projection",
        metavar="PROJTYPE",
        default="plate-carree",
        help="The projection type of the input image (default: %(default)s; choices: %(choices)s)",
        choices=[
            "plate-carree",
            "plate-carree-galactic",
            "plate-carree-ecliptic",
            "plate-carree-planet",
            "plate-carree-planet-zeroleft",
            "plate-carree-planet-zeroright",
            "plate-carree-panorama",
        ],
    )
    parser.add_argument(
        "--parallelism",
        "-j",
        metavar="COUNT",
        type=int,
        help="The parallelization level (default: use all CPUs if OS supports; specify `1` to force serial processing)",
    )
    parser.add_argument(
        "imgpath",
        metavar="PATH",
        help="The image file to be tiled",
    )
    parser.add_argument(
        "depth",
        metavar="DEPTH",
        type=int,
        help="The depth of the TOAST layer to sample",
    )


def tile_allsky_impl(settings):
    from .builder import Builder
    from .image import ImageLoader
    from .pyramid import PyramidIO

    img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
    pio = PyramidIO(settings.outdir)
    is_planet = False
    is_pano = False

    if settings.projection == "plate-carree":
        from .samplers import plate_carree_sampler

        sampler = plate_carree_sampler(img.asarray())
    elif settings.projection == "plate-carree-galactic":
        from .samplers import plate_carree_galactic_sampler

        sampler = plate_carree_galactic_sampler(img.asarray())
    elif settings.projection == "plate-carree-ecliptic":
        from .samplers import plate_carree_ecliptic_sampler

        sampler = plate_carree_ecliptic_sampler(img.asarray())
    elif settings.projection == "plate-carree-planet":
        from .samplers import plate_carree_planet_sampler

        sampler = plate_carree_planet_sampler(img.asarray())
        is_planet = True
    elif settings.projection == "plate-carree-planet-zeroleft":
        from .samplers import plate_carree_planet_zeroleft_sampler

        sampler = plate_carree_planet_zeroleft_sampler(img.asarray())
        is_planet = True
    elif settings.projection == "plate-carree-planet-zeroright":
        from .samplers import plate_carree_zeroright_sampler

        sampler = plate_carree_zeroright_sampler(img.asarray())
        is_planet = True
    elif settings.projection == "plate-carree-panorama":
        from .samplers import plate_carree_sampler

        sampler = plate_carree_sampler(img.asarray())
        is_pano = True
    else:
        die(
            "the image projection type {!r} is not recognized".format(
                settings.projection
            )
        )

    builder = Builder(pio)

    # Do the thumbnail first since for large inputs it can be the memory high-water mark!
    if settings.placeholder_thumbnail:
        builder.make_placeholder_thumbnail()
    else:
        builder.make_thumbnail_from_other(img)

    builder.toast_base(
        sampler,
        settings.depth,
        is_planet=is_planet,
        is_pano=is_pano,
        parallel=settings.parallelism,
        cli_progress=True,
    )
    builder.set_name(settings.name)
    builder.write_index_rel_wtml()

    print(
        f'Successfully tiled input "{settings.imgpath}" at level {builder.imgset.tile_levels}.'
    )
    print("To create parent tiles, consider running:")
    print()
    print(f"   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}")


# "tile_healpix" subcommand


def tile_healpix_getparser(parser):
    parser.add_argument(
        "--outdir",
        metavar="PATH",
        default=".",
        help="The root directory of the output tile pyramid",
    )
    parser.add_argument(
        "fitspath",
        metavar="PATH",
        help="The HEALPix FITS file to be tiled",
    )
    parser.add_argument(
        "depth",
        metavar="DEPTH",
        type=int,
        help="The depth of the TOAST layer to sample",
    )


def tile_healpix_impl(settings):
    from .builder import Builder
    from .pyramid import PyramidIO
    from .samplers import healpix_fits_file_sampler

    pio = PyramidIO(settings.outdir, default_format="npy")
    sampler = healpix_fits_file_sampler(settings.fitspath)
    builder = Builder(pio)
    builder.toast_base(sampler, settings.depth)
    builder.write_index_rel_wtml()

    print(
        f'Successfully tiled input "{settings.fitspath}" at level {builder.imgset.tile_levels}.'
    )
    print("To create parent tiles, consider running:")
    print()
    print(f"   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}")


# "tile_multi_tan" subcommand


def tile_multi_tan_getparser(parser):
    parser.add_argument(
        "--parallelism",
        "-j",
        metavar="COUNT",
        type=int,
        help="The parallelization level (default: use all CPUs; specify `1` to force serial processing)",
    )
    parser.add_argument(
        "--hdu-index",
        metavar="INDEX",
        type=int,
        default=0,
        help="Which HDU to load in each input FITS file",
    )
    parser.add_argument(
        "--outdir",
        metavar="PATH",
        default=".",
        help="The root directory of the output tile pyramid",
    )
    parser.add_argument(
        "paths",
        metavar="PATHS",
        action=EnsureGlobsExpandedAction,
        nargs="+",
        help="The FITS files with image data",
    )


def tile_multi_tan_impl(settings):
    from .builder import Builder
    from .collection import SimpleFitsCollection
    from .multi_tan import MultiTanProcessor
    from .pyramid import PyramidIO

    pio = PyramidIO(settings.outdir, default_format="fits")
    builder = Builder(pio)

    collection = SimpleFitsCollection(settings.paths, hdu_index=settings.hdu_index)

    mtp = MultiTanProcessor(collection)
    mtp.compute_global_pixelization(builder)
    mtp.tile(
        pio,
        parallel=settings.parallelism,
        cli_progress=True,
    )
    builder.write_index_rel_wtml()

    print(f"Successfully tiled inputs at level {builder.imgset.tile_levels}.")
    print("To create parent tiles, consider running:")
    print()
    print(f"   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}")


# "tile_study" subcommand


def tile_study_getparser(parser):
    from .image import ImageLoader

    ImageLoader.add_arguments(parser)

    parser.add_argument(
        "--name",
        metavar="NAME",
        default="Toasty",
        help="The image name to embed in the output WTML file (default: %(default)s)",
    )
    parser.add_argument(
        "--avm",
        action="store_true",
        help="Expect the input image to have AVM positioning tags",
    )
    parser.add_argument(
        "--fits-wcs",
        metavar="PATH",
        help="Get WCS information from this FITS file",
    )
    parser.add_argument(
        "--placeholder-thumbnail",
        action="store_true",
        help="Do not attempt to thumbnail the input image -- saves memory for large inputs",
    )
    parser.add_argument(
        "--outdir",
        metavar="PATH",
        default=".",
        help="The root directory of the output tile pyramid",
    )
    parser.add_argument(
        "imgpath",
        metavar="PATH",
        help="The study image file to be tiled",
    )


def tile_study_impl(settings):
    from .builder import Builder
    from .image import ImageLoader
    from .pyramid import PyramidIO

    img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
    pio = PyramidIO(settings.outdir, default_format=img.default_format)
    builder = Builder(pio)

    # First, prepare the tiling, because we can't correctly apply WCS until we
    # know the details of the tiling parameters

    tiling = builder.prepare_study_tiling(img)

    # Now deal with the WCS

    if settings.avm:
        # We don't *always* check for AVM because pyavm prints out a lot of junk
        # and may not be installed.
        try:
            from pyavm import AVM
        except ImportError:
            die("cannot use AVM data: you must install the `pyavm` package")

        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                avm = AVM.from_image(settings.imgpath)
        except Exception:
            print(
                f"error: failed to read AVM tags of input `{settings.imgpath}`",
                file=sys.stderr,
            )
            raise

        builder.apply_avm_info(avm, img.width, img.height)
    elif settings.fits_wcs:
        try:
            from astropy.io import fits
            from astropy.wcs import WCS
        except ImportError:
            die("cannot use FITS WCS: you must install the `astropy` package")

        try:
            with fits.open(settings.fits_wcs) as hdul:
                # TODO: more flexibility in HDU choice:
                hdu = hdul[0]
                wcs = WCS(hdu, fobj=hdul)
        except Exception as e:
            die(f"cannot read WCS from FITS file `{settings.fits_wcs}`: {e}")

        wcs_width = hdu.header.get("IMAGEW")  # Astrometry.net
        wcs_height = hdu.header.get("IMAGEH")

        if wcs_width is None and len(hdu.shape) > 1:
            wcs_width = hdu.shape[-1]
            wcs_height = hdu.shape[-2]

        if wcs_width is None:
            warn(
                f"cannot infer image dimensions from WCS FITS file `{settings.fits_wcs}`; "
                f"unable to check consistency with input image"
            )
        elif img.shape[:2] != (wcs_height, wcs_width):
            warn(
                f"image `{settings.imgpath}` has shape {img.shape}, but "
                f"WCS reference file `{settings.fits_wcs}` has shape ({wcs_height}, {wcs_width}); "
                f"astrometry may not transfer correctly"
            )

        img._wcs = wcs  # <= hack alert, but I think this is OK
        img.ensure_negative_parity()
        builder.apply_wcs_info(img.wcs, img.width, img.height)
    elif img.wcs is not None:
        # Study images must have negative parity.
        img.ensure_negative_parity()
        builder.apply_wcs_info(img.wcs, img.width, img.height)
    else:
        builder.default_tiled_study_astrometry()

    # Do the thumbnail before the main image since for large inputs it can be
    # the memory high-water mark!
    if settings.placeholder_thumbnail:
        builder.make_placeholder_thumbnail()
    else:
        builder.make_thumbnail_from_other(img)

    # Finally, actually tile the image
    builder.execute_study_tiling(img, tiling, cli_progress=True)

    # Final metadata
    builder.set_name(settings.name)
    builder.write_index_rel_wtml()

    print(
        f'Successfully tiled input "{settings.imgpath}" at level {builder.imgset.tile_levels}.'
    )
    print("To create parent tiles, consider running:")
    print()
    print(f"   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}")


# "tile_wwtl" subcommand


def tile_wwtl_getparser(parser):
    from .image import ImageLoader

    ImageLoader.add_arguments(parser)

    parser.add_argument(
        "--name",
        metavar="NAME",
        default="Toasty",
        help="The image name to embed in the output WTML file (default: %(default)s)",
    )
    parser.add_argument(
        "--placeholder-thumbnail",
        action="store_true",
        help="Do not attempt to thumbnail the input image -- saves memory for large inputs",
    )
    parser.add_argument(
        "--outdir",
        metavar="PATH",
        default=".",
        help="The root directory of the output tile pyramid",
    )
    parser.add_argument(
        "wwtl_path",
        metavar="WWTL-PATH",
        help="The WWTL layer file to be processed",
    )


def tile_wwtl_impl(settings):
    from .builder import Builder
    from .pyramid import PyramidIO

    pio = PyramidIO(settings.outdir)
    builder = Builder(pio)
    img = builder.load_from_wwtl(settings, settings.wwtl_path, cli_progress=True)

    # Do the thumbnail first since for large inputs it can be the memory high-water mark!
    if settings.placeholder_thumbnail:
        builder.make_placeholder_thumbnail()
    else:
        builder.make_thumbnail_from_other(img)

    builder.set_name(settings.name)
    builder.write_index_rel_wtml()

    print(
        f'Successfully tiled input "{settings.wwtl_path}" at level {builder.imgset.tile_levels}.'
    )
    print("To create parent tiles, consider running:")
    print()
    print(f"   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}")


# "transform" subcommand


def transform_getparser(parser):
    subparsers = parser.add_subparsers(dest="transform_command")

    parser = subparsers.add_parser("fx3-to-rgb")
    parser.add_argument(
        "--parallelism",
        "-j",
        metavar="COUNT",
        type=int,
        help="The parallelization level (default: use all CPUs; specify `1` to force serial processing)",
    )
    parser.add_argument(
        "--start",
        metavar="DEPTH",
        type=int,
        help="How deep in the tile pyramid to proceed wth the transformation",
    )
    parser.add_argument(
        "--clip",
        "-c",
        metavar="NUMBER",
        type=float,
        default=1.0,
        help="The level at which to start flipping the floating-point data",
    )
    parser.add_argument(
        "--outdir",
        metavar="PATH",
        help="Output transformed data into this directory, instead of operating in-place",
    )
    parser.add_argument(
        "pyramid_dir",
        metavar="DIR",
        help="The directory containing the tile pyramid to transform",
    )

    parser = subparsers.add_parser("u8-to-rgb")
    parser.add_argument(
        "--parallelism",
        "-j",
        metavar="COUNT",
        type=int,
        help="The parallelization level (default: use all CPUs; specify `1` to force serial processing)",
    )
    parser.add_argument(
        "--start",
        metavar="DEPTH",
        type=int,
        help="How deep in the tile pyramid to proceed wth the transformation",
    )
    parser.add_argument(
        "--outdir",
        metavar="PATH",
        help="Output transformed data into this directory, instead of operating in-place",
    )
    parser.add_argument(
        "pyramid_dir",
        metavar="DIR",
        help="The directory containing the tile pyramid to transform",
    )


def transform_impl(settings):
    from .pyramid import PyramidIO

    if settings.transform_command is None:
        print('Run the "transform" command with `--help` for help on its subcommands')
        return

    if settings.transform_command == "fx3-to-rgb":
        pio = PyramidIO(settings.pyramid_dir)
        pio_out = None

        if settings.outdir is not None:
            pio_out = PyramidIO(settings.outdir)

        from .transform import f16x3_to_rgb

        f16x3_to_rgb(
            pio,
            settings.start,
            clip=settings.clip,
            pio_out=pio_out,
            parallel=settings.parallelism,
            cli_progress=True,
        )
    elif settings.transform_command == "u8-to-rgb":
        pio = PyramidIO(settings.pyramid_dir)
        pio_out = None

        if settings.outdir is not None:
            pio_out = PyramidIO(settings.outdir)

        from .transform import u8_to_rgb

        u8_to_rgb(
            pio,
            settings.start,
            pio_out=pio_out,
            parallel=settings.parallelism,
            cli_progress=True,
        )
    else:
        die('unrecognized "transform" subcommand ' + settings.transform_command)


# "view" subcommand
#
# This largely duplicates `wwtdatatool preview` in the `wwt_data_formats`
# package. The difference is that we automagically tile and output the
# `index_rel.wtml`, while `wwtdatatool` requires it to be preexisting.


def view_getparser(parser):
    from .collection import CollectionLoader

    CollectionLoader.add_arguments(parser)

    parser.add_argument(
        "--parallelism",
        "-j",
        metavar="COUNT",
        type=int,
        help="The parallelization level (default: use all CPUs; specify `1` to force serial processing)",
    )
    parser.add_argument(
        "--browser",
        "-b",
        metavar="BROWSER-TYPE",
        help="The type of browser to use for the viewer (as per Python webbrowser)",
    )
    parser.add_argument(
        "--appurl",
        metavar="URL",
        help="The URL of the app to use (intended for developers)",
    )
    parser.add_argument(
        "paths",
        metavar="PATHS",
        action=EnsureGlobsExpandedAction,
        nargs="+",
        help="The FITS file(s) with image data",
    )


def view_impl(settings):
    from wwt_data_formats.server import preview_wtml
    from .collection import CollectionLoader
    from .fits_tiler import FitsTiler

    coll = CollectionLoader.create_from_args(settings).load_paths(settings.paths)

    # Ignore any astropy WCS/FITS warnings, which can spew a lot of annoying output.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        tiler = FitsTiler(coll)
        tiler.tile(cli_progress=True)

    preview_wtml(
        os.path.join(tiler.out_dir, "index_rel.wtml"),
        browser=settings.browser,
        app_type="research",
        app_url=settings.appurl,
    )


# The CLI driver:


def entrypoint(args=None):
    """The entrypoint for the \"toasty\" command-line interface.

    Parameters
    ----------
    args : iterable of str, or None (the default)
      The arguments on the command line. The first argument should be
      a subcommand name or global option; there is no ``argv[0]``
      parameter.

    """
    # Set up the subcommands. We use locals() and globals() in a fairly gross
    # way to avoid circular import issues in Sphinx.

    from .pipeline.cli import pipeline_getparser, pipeline_impl

    names = dict(locals())
    names.update(globals())

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subcommand")
    commands = set()

    for py_name, value in names.items():
        if py_name.endswith("_getparser"):
            cmd_name = py_name[:-10].replace("_", "-")
            subparser = subparsers.add_parser(cmd_name)
            value(subparser)
            commands.add(cmd_name)

    # What did we get?

    settings = parser.parse_args(args)

    if settings.subcommand is None:
        print("Run me with --help for help. Allowed subcommands are:")
        print()
        for cmd in sorted(commands):
            print("   ", cmd)
        return

    py_name = settings.subcommand.replace("-", "_")

    impl = names.get(py_name + "_impl")
    if impl is None:
        die('no such subcommand "{}"'.format(settings.subcommand))

    # OK to go!

    impl(settings)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2021 the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import pytest
from . import test_path
from .. import collection

try:
    from astropy.io import fits
    HAS_ASTRO = True
except ImportError:
    HAS_ASTRO = False


class TestCollection(object):

    @pytest.mark.skipif('not HAS_ASTRO')
    def test_is_multi_tan(self):
        coll = collection.SimpleFitsCollection([test_path('wcs512.fits.gz')])
        assert coll._is_multi_tan()

        coll = collection.SimpleFitsCollection([test_path('herschel_spire.fits.gz'),
                                                test_path('herschel_spire.fits.gz')])
        assert coll._is_multi_tan()

        coll = collection.SimpleFitsCollection([test_path('wcs512.fits.gz'),
                                                test_path('herschel_spire.fits.gz')])
        assert not coll._is_multi_tan()



# -*- mode: python; coding: utf-8 -*-
# Copyright 2022 the AAS WorldWide Telescope project
# Licensed under the MIT License.

import numpy as np
import pytest

from ..image import Image, ImageMode


class ModeInfo(object):
    def __init__(self, mode, dtype, demo_shape, default_format):
        self.mode = mode
        self.dtype = dtype
        self.demo_shape = demo_shape
        self.default_format = default_format


MODE_INFO = [
    ModeInfo(ImageMode.RGB, np.uint8, (10, 10, 3), "png"),
    ModeInfo(ImageMode.RGBA, np.uint8, (10, 10, 4), "png"),
    ModeInfo(ImageMode.F32, np.float32, (10, 10), "npy"),
    ModeInfo(ImageMode.F64, np.float64, (10, 10), "npy"),
    ModeInfo(ImageMode.F16x3, np.float16, (10, 10, 3), "npy"),
    ModeInfo(ImageMode.U8, np.uint8, (10, 10), "npy"),
    ModeInfo(ImageMode.I32, np.int32, (10, 10), "npy"),
]


def test_from_array_info():
    for mi in MODE_INFO:
        assert ImageMode.from_array_info(mi.demo_shape, mi.dtype) == mi.mode

    with pytest.raises(ValueError):
        s = (10, 10)
        ImageMode.from_array_info(s, np.complex64)

        s = (10, 10, 3)
        ImageMode.from_array_info(s, np.float32)

        s = (10, 10, 4)
        ImageMode.from_array_info(s, np.int8)


def test_image_interface():
    for mi in MODE_INFO:
        arr = np.zeros(mi.demo_shape, dtype=mi.dtype)
        img = Image.from_array(arr)
        img._default_format = None
        assert img.default_format == mi.default_format
        mb = mi.mode.make_maskable_buffer(8, 8)
        img.fill_into_maskable_buffer(
            mb, slice(1, 5), slice(1, 5), slice(4, 8), slice(4, 8)
        )

        mb.asarray().fill(1)
        img.update_into_maskable_buffer(
            mb, slice(1, 5), slice(1, 5), slice(4, 8), slice(4, 8)
        )
        assert mb.asarray().flat[0] == 1

        img.clear()



# -*- mode: python; coding: utf-8 -*-
# Copyright 2013-2022 Chris Beaumont and the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import os
from tempfile import mkstemp, mkdtemp
from shutil import rmtree

import pytest
import numpy as np

try:
    import healpy as hp
    from astropy.io import fits

    HAS_ASTRO = True
except ImportError:
    HAS_ASTRO = False

try:
    import OpenEXR

    HAS_OPENEXR = True
except ImportError:
    HAS_OPENEXR = False

from . import test_path
from .. import cli, toast
from ..image import ImageLoader
from ..jpeg2000 import ChunkedJPEG2000Reader, HAS_JPEG2000
from ..pyramid import Pos, PyramidIO
from ..toast import sample_layer, sample_layer_filtered
from ..transform import f16x3_to_rgb


def test_mid():
    from .._libtoasty import mid

    result = mid((0, 0), (np.pi / 2, 0))
    expected = np.pi / 4, 0
    np.testing.assert_array_almost_equal(result, expected)

    result = mid((0, 0), (0, 1))
    expected = 0, 0.5
    np.testing.assert_array_almost_equal(result, expected)


def test_area():
    MAX_DEPTH = 6
    areas = {}

    for t in toast.generate_tiles(MAX_DEPTH, bottom_only=False):
        a = areas.get(t.pos.n, 0)
        areas[t.pos.n] = a + toast.toast_tile_area(t)

    for d in range(1, MAX_DEPTH + 1):
        np.testing.assert_almost_equal(areas[d], 4 * np.pi)


def test_tile_for_point_boundaries():
    # These test points occur at large-scale tile boundaries and so we're not
    # picky about where they land in the tiling -- either tile on a border is
    # OK.

    from ..toast import toast_tile_for_point

    latlons = [
        (0.0, 0.0),
        (0.0, -0.5 * np.pi),
        (0.0, 0.5 * np.pi),
        (0.0, np.pi),
        (0.0, 1.5 * np.pi),
        (0.0, 2 * np.pi),
        (0.0, 2.5 * np.pi),
        (-0.5 * np.pi, 0.0),
        (-0.5 * np.pi, -np.pi),
        (-0.5 * np.pi, np.pi),
        (0.5 * np.pi, 0.0),
        (0.5 * np.pi, -np.pi),
        (0.5 * np.pi, np.pi),
    ]

    for depth in range(4):
        for lat, lon in latlons:
            tile = toast_tile_for_point(depth, lat, lon)


def test_tile_for_point_specifics():
    from ..toast import toast_tile_for_point

    test_data = {
        (0.1, 0.1): [
            (0, 0, 0),
            (1, 1, 0),
            (2, 3, 1),
            (3, 7, 3),
            (4, 14, 7),
            (5, 29, 14),
            (6, 59, 29),
            (7, 119, 59),
            (8, 239, 119),
            (9, 479, 239),
            (10, 959, 479),
            (11, 1918, 959),
            (12, 3837, 1918),
        ]
    }

    for (lat, lon), nxys in test_data.items():
        for nxy in nxys:
            tile = toast_tile_for_point(nxy[0], lat, lon)
            assert tile.pos.n == nxy[0]
            assert tile.pos.x == nxy[1]
            assert tile.pos.y == nxy[2]


def test_pixel_for_point():
    from ..toast import toast_pixel_for_point

    test_data = {
        (0.1, 0.1): (3, 7, 3, 126, 191),
        (-0.4, 2.2): (3, 1, 1, 200, 75),
    }

    for (lat, lon), (n, x, y, px, py) in test_data.items():
        shallow_tile, frac_x, frac_y = toast_pixel_for_point(n, lat, lon)
        assert shallow_tile.pos.n == n
        assert shallow_tile.pos.x == x
        assert shallow_tile.pos.y == y
        assert px == int(round(frac_x))
        assert py == int(round(frac_y))
        deep_tile, _, _ = toast_pixel_for_point(n + 8, lat, lon)
        assert deep_tile.pos.x % 256 == px
        assert deep_tile.pos.y % 256 == py


def image_test(expected, actual, err_msg):
    resid = np.abs(1.0 * actual - expected)
    if np.median(resid) < 10:
        return

    _, pth = mkstemp(suffix=".png")
    import PIL.Image

    PIL.Image.fromarray(np.hstack((expected, actual))).save(pth)
    pytest.fail("%s. MedResid=%f. Saved to %s" % (err_msg, np.median(resid), pth))


class PyramidTester(object):
    def setup_method(self, _method):
        self.work_dir = mkdtemp()
        self.pio = PyramidIO(self.work_dir)

    def teardown_method(self, _method):
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    def verify_level1(
        self,
        ref="earth_toasted_sky",
        format=None,
        expected_2d=False,
        drop_alpha=False,
        planetary=False,
    ):
        for n, x, y in [(1, 0, 0), (1, 0, 1), (1, 1, 0), (1, 1, 1)]:
            if planetary:
                # to transform sky to planetary, we need to mirror the global
                # image horizontally. So here we mirror on a tile level, and
                # below we mirror on a pixel level.
                ref_x = 1 - x
                warn = f"; this is mirrored from input position ({n},{x},{y}) for planetary mode"
            else:
                ref_x = x
                warn = ""

            ref_path = test_path(ref, str(n), str(y), "%i_%i.png" % (y, ref_x))
            expected = ImageLoader().load_path(ref_path).asarray()

            if planetary:
                expected = expected[:, ::-1]

            pos = Pos(n=n, x=x, y=y)
            observed = self.pio.read_image(pos, format=format).asarray()

            if expected_2d:
                expected = expected.mean(axis=2)

            if drop_alpha:
                assert observed.shape[2] == 4
                observed = observed[..., :3]

            image_test(expected, observed, "Failed for %s%s" % (ref_path, warn))


class TestSampleLayer(PyramidTester):
    def test_plate_carree(self):
        from ..samplers import plate_carree_sampler

        img = ImageLoader().load_path(
            test_path("Equirectangular_projection_SW-tweaked.jpg")
        )
        sampler = plate_carree_sampler(img.asarray())
        sample_layer(self.pio, sampler, 1, format="png")
        self.verify_level1()

    def test_plate_carree_ecliptic(self):
        from ..samplers import plate_carree_ecliptic_sampler

        img = ImageLoader().load_path(test_path("tess_platecarree_ecliptic_512.jpg"))
        sampler = plate_carree_ecliptic_sampler(img.asarray())
        sample_layer(self.pio, sampler, 1, format="png")
        self.verify_level1(ref="tess")

    @pytest.mark.skipif("not HAS_OPENEXR")
    def test_earth_plate_carree_exr(self):
        from ..samplers import plate_carree_sampler

        img = ImageLoader().load_path(
            test_path("Equirectangular_projection_SW-tweaked.exr")
        )
        sampler = plate_carree_sampler(img.asarray())
        sample_layer(self.pio, sampler, 1, format="npy")
        f16x3_to_rgb(self.pio, 1, parallel=1, out_format="png")
        self.verify_level1()

    @pytest.mark.skipif("not HAS_JPEG2000")
    def test_earth_plate_carree_jpeg2000_chunked_planetary(self):
        from ..samplers import ChunkedPlateCarreeSampler
        from ..toast import ToastCoordinateSystem

        img = ChunkedJPEG2000Reader(
            test_path("Equirectangular_projection_SW-tweaked.jp2")
        )
        # this currently (2021 Oct) only supports planetary coordinates:
        chunker = ChunkedPlateCarreeSampler(img, planetary=True)

        for ichunk in range(chunker.n_chunks):
            tile_filter = chunker.filter(ichunk)
            sampler = chunker.sampler(ichunk)
            sample_layer_filtered(
                self.pio,
                tile_filter,
                sampler,
                1,  # depth
                coordsys=ToastCoordinateSystem.PLANETARY,
                parallel=1,
            )

        self.verify_level1(drop_alpha=True, planetary=True)

    @pytest.mark.skipif("not HAS_ASTRO")
    def test_healpix_equ(self):
        from ..samplers import healpix_fits_file_sampler

        sampler = healpix_fits_file_sampler(test_path("earth_healpix_equ.fits"))
        sample_layer(self.pio, sampler, 1, format="npy")
        self.verify_level1(format="npy", expected_2d=True)

    @pytest.mark.skipif("not HAS_ASTRO")
    def test_healpix_gal(self):
        from ..samplers import healpix_fits_file_sampler

        sampler = healpix_fits_file_sampler(test_path("earth_healpix_gal.fits"))
        sample_layer(self.pio, sampler, 1, format="fits")
        self.verify_level1(format="fits", expected_2d=True)


class TestCliBasic(PyramidTester):
    """
    CLI tests.
    """

    def inner_test(self, image_path, projection):
        args = [
            "tile-allsky",
            "--name=Earth",
            f"--projection={projection}",
            "--outdir",
            self.work_path(),
            image_path,
            "2",
        ]
        cli.entrypoint(args)
        cli.entrypoint(["cascade", "--start", "2", self.work_path()])
        self.verify_level1(planetary=True, drop_alpha=True)

    def test_planet(self):
        self.inner_test(
            test_path("Equirectangular_projection_SW-tweaked.jpg"),
            "plate-carree-planet",
        )

    def test_planet_zeroleft(self):
        """
        Our source image is a standard planetary map where longitudes range from
        -180 to 180, left to right. So to get the correct zeroleft input, we
        need to swap the left and right halves.
        """
        img = ImageLoader().load_path(
            test_path("Equirectangular_projection_SW-tweaked.jpg")
        )
        hw = img.width // 2
        buf = img.asarray()
        temp = buf[:, hw:].copy()
        buf[:, hw:] = buf[:, :hw]
        buf[:, :hw] = temp
        img._array = buf  # hack!
        img._pil = None
        p = self.work_path("zeroleft.jpg")
        img.save(p, format="jpg")
        self.inner_test(
            p,
            "plate-carree-planet-zeroleft",
        )

    def test_planet_zeroright(self):
        """
        The horizontal mirror of zeroleft. Longitudes increase to the left, but
        the right edge of the image is longitude 0.
        """
        img = ImageLoader().load_path(
            test_path("Equirectangular_projection_SW-tweaked.jpg")
        )
        hw = img.width // 2
        buf = img.asarray()
        temp = buf[:, hw:].copy()
        buf[:, hw:] = buf[:, :hw]
        buf[:, :hw] = temp
        img._array = buf[:, ::-1]  # hack!
        img._pil = None
        p = self.work_path("zeroright.jpg")
        img.save(p, format="jpg")
        self.inner_test(
            p,
            "plate-carree-planet-zeroright",
        )



# -*- mode: python; coding: utf-8 -*-
# Copyright 2013-2020 Chris Beaumont and the AAS WorldWide Telescope team
# Licensed under the MIT License

from __future__ import absolute_import, division, print_function

from Cython.Distutils import build_ext  # in pyproject.toml
import numpy as np  # in pyproject.toml
import os
from setuptools import setup, Extension


def get_long_desc():
    in_preamble = True
    lines = []

    with open("README.md", "rt", encoding="utf8") as f:
        for line in f:
            if in_preamble:
                if line.startswith("<!--pypi-begin-->"):
                    in_preamble = False
            else:
                if line.startswith("<!--pypi-end-->"):
                    break
                else:
                    lines.append(line)

    lines.append(
        """

For more information, including installation instructions, please visit [the
project homepage].

[the project homepage]: https://toasty.readthedocs.io/
"""
    )
    return "".join(lines)


setup_args = dict(
    name="toasty",  # cranko project-name
    version="0.16.1",  # cranko project-version
    description="Generate TOAST image tile pyramids from existing image data",
    long_description=get_long_desc(),
    long_description_content_type="text/markdown",
    url="https://toasty.readthedocs.io/",
    license="MIT",
    platforms="Linux, Mac OS X",
    author="Chris Beaumont, AAS WorldWide Telescope Team",
    author_email="wwt@aas.org",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    packages=[
        "toasty",
        "toasty.pipeline",
        "toasty.tests",
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "toasty=toasty.cli:entrypoint",
        ],
    },
    install_requires=[
        "filelock>=3",
        "numpy>=1.7",
        "pillow>=7.0",
        "PyYAML>=5.0",
        "reproject",
        "shapely",
        "tqdm>=4.0",
        "wwt_data_formats>=0.12",
    ],
    extras_require={
        "test": [
            "coveralls",
            "pytest-cov",
        ],
        "docs": [
            "astropy",
            "astropy-sphinx-theme",
            "numpydoc",
            "sphinx",
            "sphinx-automodapi",
        ],
    },
    cmdclass={
        "build_ext": build_ext,
    },
    ext_modules=[
        Extension("toasty._libtoasty", ["toasty/_libtoasty.pyx"]),
    ],
    include_dirs=[
        np.get_include(),
    ],
)

for e in setup_args["ext_modules"]:
    e.cython_directives = {"language_level": "3"}

if __name__ == "__main__":
    setup(**setup_args)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2022 the AAS WorldWide Telescope project.
# Licensed under the MIT License.

"""
Entrypoints for the "toasty pipeline" command-line tools.
"""

__all__ = """
pipeline_getparser
pipeline_impl
""".split()

from fnmatch import fnmatch
import glob
import os.path
import sys
from wwt_data_formats.cli import EnsureGlobsExpandedAction

from ..cli import die
from . import NotActionableError


def evaluate_imageid_args(searchdir, args):
    """
    Figure out which image-ID's to process.
    """

    matched_ids = set()
    globs_todo = set()

    for arg in args:
        if glob.has_magic(arg):
            globs_todo.add(arg)
        else:
            # If an ID is explicitly (non-gobbily) added, always add it to the
            # list, without checking if it exists in `searchdir`. We could check
            # for it in searchdir now, but we'll have to check later anyway, so
            # we don't bother.
            matched_ids.add(arg)

    if len(globs_todo):
        for filename in os.listdir(searchdir):
            for g in globs_todo:
                if fnmatch(filename, g):
                    matched_ids.add(filename)
                    break

    return sorted(matched_ids)


# The "approve" subcommand


def approve_setup_parser(parser):
    parser.add_argument(
        "--workdir",
        metavar="PATH",
        default=".",
        help="The working directory for this processing session",
    )
    parser.add_argument(
        "cand_ids",
        nargs="+",
        action=EnsureGlobsExpandedAction,
        metavar="IMAGE-ID",
        help="Name(s) of image(s) to approve for publication (globs accepted)",
    )


def approve_impl(settings):
    from wwt_data_formats.folder import Folder, make_absolutizing_url_mutator
    from . import PipelineManager

    mgr = PipelineManager(settings.workdir)
    mgr.ensure_config()

    pub_url_prefix = mgr._config.get("publish_url_prefix")
    if pub_url_prefix:
        if pub_url_prefix[-1] != "/":
            pub_url_prefix += "/"

    proc_dir = mgr._ensure_dir("processed")
    app_dir = mgr._ensure_dir("approved")

    for cid in evaluate_imageid_args(proc_dir, settings.cand_ids):
        if not os.path.isdir(os.path.join(proc_dir, cid)):
            die(f"no such processed candidate ID {cid!r}")

        index_path = os.path.join(proc_dir, cid, "index.wtml")
        prefix = pub_url_prefix + cid + "/"

        try:
            f = Folder.from_file(os.path.join(proc_dir, cid, "index_rel.wtml"))
            f.mutate_urls(make_absolutizing_url_mutator(prefix))

            with open(index_path, "wt", encoding="utf8") as f_out:
                f.write_xml(f_out)
        except Exception as e:
            print(
                "error: failed to create index.wtml from index_rel.wtml",
                file=sys.stderr,
            )

            try:
                os.remove(index_path)
            except Exception:
                pass

            raise

        os.rename(os.path.join(proc_dir, cid), os.path.join(app_dir, cid))


# The "fetch" subcommand


def fetch_setup_parser(parser):
    parser.add_argument(
        "--workdir",
        metavar="PATH",
        default=".",
        help="The working directory for this processing session",
    )
    parser.add_argument(
        "cand_ids",
        nargs="+",
        action=EnsureGlobsExpandedAction,
        metavar="CAND-ID",
        help="Name(s) of candidate(s) to fetch and prepare for processing (globs accepted)",
    )


def fetch_impl(settings):
    from . import PipelineManager

    mgr = PipelineManager(settings.workdir)
    cand_dir = mgr._ensure_dir("candidates")
    rej_dir = mgr._ensure_dir("rejects")
    src = mgr.get_image_source()

    for cid in evaluate_imageid_args(cand_dir, settings.cand_ids):
        # Funky structure here is to try to ensure that cdata is closed in case
        # a NotActionable happens, so that we can move the directory on Windows.
        try:
            try:
                cdata = open(os.path.join(cand_dir, cid), "rb")
            except FileNotFoundError:
                die(f"no such candidate ID {cid!r}")

            try:
                print(f"fetching {cid} ... ", end="")
                sys.stdout.flush()
                cachedir = mgr._ensure_dir("cache_todo", cid)
                src.fetch_candidate(cid, cdata, cachedir)
                print("done")
            finally:
                cdata.close()
        except NotActionableError as e:
            print("not usable:", e)
            os.rename(os.path.join(cand_dir, cid), os.path.join(rej_dir, cid))
            os.rmdir(cachedir)


# The "init" subcommand


def init_setup_parser(parser):
    parser.add_argument(
        "--azure-conn-env",
        metavar="ENV-VAR-NAME",
        help="The name of an environment variable contain an Azure Storage "
        "connection string",
    )
    parser.add_argument(
        "--azure-container",
        metavar="CONTAINER-NAME",
        help="The name of a blob container in the Azure storage account",
    )
    parser.add_argument(
        "--azure-path-prefix",
        metavar="PATH-PREFIX",
        help="A slash-separated path prefix for blob I/O within the container",
    )
    parser.add_argument(
        "--local", metavar="PATH", help="Use the local-disk I/O backend"
    )
    parser.add_argument(
        "workdir",
        nargs="?",
        metavar="PATH",
        default=".",
        help="The working directory for this processing session",
    )


def _pipeline_io_from_settings(settings):
    from . import azure_io, local_io

    if settings.local:
        return local_io.LocalPipelineIo(settings.local)

    if settings.azure_conn_env:
        conn_str = os.environ.get(settings.azure_conn_env)
        if not conn_str:
            die(
                "--azure-conn-env=%s provided, but that environment variable is unset"
                % settings.azure_conn_env
            )

        if not settings.azure_container:
            die("--azure-container-name must be provided if --azure-conn-env is")

        path_prefix = settings.azure_path_prefix
        if not path_prefix:
            path_prefix = ""

        azure_io.assert_enabled()

        return azure_io.AzureBlobPipelineIo(
            conn_str, settings.azure_container, path_prefix
        )

    die("An I/O backend must be specified with the arguments --local or --azure-*")


def init_impl(settings):
    pipeio = _pipeline_io_from_settings(settings)
    os.makedirs(settings.workdir, exist_ok=True)
    pipeio.save_config(os.path.join(settings.workdir, "toasty-store-config.yaml"))


# The "refresh" subcommand
#
# TODO: for large feeds, we should potentially add features to make it so that
# we don't re-check every single candidate that's ever been posted.


def refresh_setup_parser(parser):
    parser.add_argument(
        "--workdir",
        nargs="?",
        metavar="PATH",
        default=".",
        help="The working directory for this processing session",
    )


def refresh_impl(settings):
    from . import PipelineManager

    mgr = PipelineManager(settings.workdir)
    cand_dir = mgr._ensure_dir("candidates")
    rej_dir = mgr._ensure_dir("rejects")
    src = mgr.get_image_source()
    n_cand = 0
    n_saved = 0
    n_done = 0
    n_skipped = 0
    n_rejected = 0

    for cand in src.query_candidates():
        n_cand += 1
        uniq_id = cand.get_unique_id()

        if mgr._pipeio.check_exists(uniq_id, "index.wtml"):
            n_done += 1
            continue  # skip already-done inputs

        if mgr._pipeio.check_exists(uniq_id, "skip.flag"):
            n_skipped += 1
            continue  # skip inputs that are explicitly flagged

        cand_path = os.path.join(cand_dir, uniq_id)

        try:
            with open(cand_path, "wb") as f:
                cand.save(f)
            n_saved += 1
        except NotActionableError as e:
            os.remove(cand_path)

            with open(os.path.join(rej_dir, uniq_id, "wb")) as f:
                pass  # for now, just touch the file

            n_rejected += 1

    print(f"analyzed {n_cand} candidates from the image source")
    print(f"  - {n_saved} processing candidates saved")
    print(f"  - {n_rejected} rejected as definitely unusable")
    print(f"  - {n_done} were already done")
    print(f"  - {n_skipped} were already marked to be ignored")
    print()
    print("See the `candidates` directory for candidate image IDs.")


# Other subcommands not yet split out.


def pipeline_getparser(parser):
    subparsers = parser.add_subparsers(dest="pipeline_command")

    def add_manager_command(name):
        subp = subparsers.add_parser(name)
        subp.add_argument(
            "--workdir",
            nargs="?",
            metavar="WORKDIR",
            default=".",
            help="The local working directory",
        )
        return subp

    approve_setup_parser(subparsers.add_parser("approve"))
    fetch_setup_parser(subparsers.add_parser("fetch"))
    add_manager_command("ignore-rejects")
    init_setup_parser(subparsers.add_parser("init"))
    add_manager_command("process-todos")
    add_manager_command("publish")
    refresh_setup_parser(subparsers.add_parser("refresh"))


def pipeline_impl(settings):
    from . import PipelineManager

    if settings.pipeline_command is None:
        print('Run the "pipeline" command with `--help` for help on its subcommands')
        return

    if settings.pipeline_command == "approve":
        approve_impl(settings)
    elif settings.pipeline_command == "fetch":
        fetch_impl(settings)
    elif settings.pipeline_command == "ignore-rejects":
        mgr = PipelineManager(settings.workdir)
        mgr.ignore_rejects()
    elif settings.pipeline_command == "init":
        init_impl(settings)
    elif settings.pipeline_command == "process-todos":
        mgr = PipelineManager(settings.workdir)
        mgr.process_todos()
    elif settings.pipeline_command == "publish":
        mgr = PipelineManager(settings.workdir)
        mgr.publish()
    elif settings.pipeline_command == "refresh":
        refresh_impl(settings)
    else:
        die('unrecognized "pipeline" subcommand ' + settings.pipeline_command)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020-2022 the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import numpy as np
import os.path
import pytest
import sys

from . import HAS_AVM, assert_xml_elements_equal, test_path
from .. import cli
from .. import study


class TestStudy(object):
    WTML = """<?xml version='1.0' encoding='UTF-8'?>
<Folder
    Browseable="True"
    Group="Explorer"
    Name="Toasty"
    Searchable="True"
>
    <Place
        Angle="0"
        AngularSize="0"
        DataSetType="Sky"
        Dec="0"
        Magnitude="0"
        Name="Toasty"
        Opacity="100"
        RA="0"
        Rotation="0"
        Thumbnail="thumb.jpg"
        ZoomLevel="1"
    >
        <ForegroundImageSet>
            <ImageSet
                BandPass="Visible"
                BaseDegreesPerTile="1"
                BaseTileLevel="0"
                BottomsUp="False"
                CenterX="0"
                CenterY="0"
                DataSetType="Sky"
                ElevationModel="False"
                FileType=".png"
                Generic="False"
                Name="Toasty"
                Projection="Tan"
                QuadTreeMap=""
                Rotation="0"
                Sparse="True"
                StockSet="False"
                TileLevels="4"
                Url="{1}/{3}/{3}_{2}.png"
                WidthFactor="2"
            >
                <ThumbnailUrl>thumb.jpg</ThumbnailUrl>
            </ImageSet>
        </ForegroundImageSet>
    </Place>
</Folder>
"""

    def setup_method(self, method):
        from tempfile import mkdtemp

        self.work_dir = mkdtemp()

    def teardown_method(self, method):
        from shutil import rmtree

        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    def test_basic(self):
        tiling = study.StudyTiling(2048, 2048)
        assert tiling._width == 2048
        assert tiling._height == 2048
        assert tiling._p2n == 2048
        assert tiling._tile_size == 8
        assert tiling._tile_levels == 3
        assert tiling._img_gx0 == 0
        assert tiling._img_gy0 == 0

    def test_preconditions(self):
        with pytest.raises(ValueError):
            study.StudyTiling(0, 1)

        with pytest.raises(ValueError):
            study.StudyTiling(1, -1)

        with pytest.raises(ValueError):
            study.StudyTiling(1, np.nan)

    def test_image_to_tile(self):
        tiling = study.StudyTiling(514, 514)
        assert tiling._p2n == 1024
        assert tiling.image_to_tile(0, 0) == (0, 0, 255, 255)
        assert tiling.image_to_tile(0, 513) == (0, 3, 255, 0)
        assert tiling.image_to_tile(513, 0) == (3, 0, 0, 255)
        assert tiling.image_to_tile(513, 513) == (3, 3, 0, 0)

    def test_sample_cli(self):
        from xml.etree import ElementTree as etree

        expected = etree.fromstring(self.WTML)

        for variants in ([], ["--placeholder-thumbnail"]):
            args = ["tile-study"]
            args += variants
            args += ["--outdir", self.work_path(), test_path("NGC253ALMA.jpg")]
            cli.entrypoint(args)

        with open(self.work_path("index_rel.wtml"), "rt", encoding="utf8") as f:
            observed = etree.fromstring(f.read())

        assert_xml_elements_equal(observed, expected)

    AVM_WTML = """<?xml version='1.0' encoding='UTF-8'?>
<Folder Browseable="True" Group="Explorer" Name="Toasty" Searchable="True">
    <Place
        Angle="0"
        AngularSize="0"
        DataSetType="Sky"
        Dec="-42.58752472831171"
        Magnitude="0"
        Name="Toasty"
        Opacity="100"
        RA="23.269985215493794"
        Rotation="0"
        Thumbnail="thumb.jpg"
        ZoomLevel="0.3370990847923382"
    >
        <ForegroundImageSet>
            <ImageSet
                BandPass="Visible"
                BaseDegreesPerTile="0.07726507807936124"
                BaseTileLevel="0"
                BottomsUp="False"
                CenterX="349.049780614"
                CenterY="-42.5874939584"
                DataSetType="Sky"
                ElevationModel="False"
                FileType=".png"
                Generic="False"
                Name="Toasty"
                OffsetX="1.886354445296938e-05"
                OffsetY="2.4372703865669735e-05"
                Projection="Tan"
                QuadTreeMap=""
                Rotation="-138.99999999999943"
                Sparse="True"
                StockSet="False"
                TileLevels="3"
                Url="{1}/{3}/{3}_{2}.png"
                WidthFactor="2"
            >
                <Credits>International Gemini Observatory/NOIRLab/NSF/AURA/B. Reynolds (Sutherland Shire Christian School)/T. Rector (University of Alaska, Anchorage)/Australian Gemini Office.</Credits>
                <CreditsUrl>https://noirlab.edu/public/images/geminiann11015a/</CreditsUrl>
                <Description>Gemini GMOS image of the barred spiral galaxy NGC 7552. Benjamin Reynolds, a 10th grade student at Sutherland Shire Christian School, suggested this target for Australia’s 2011 Gemini School Astronomy Contest and won. The picture consists of separate images taken with different filters: H-alpha (red), g (blue), r (green), and i (yellow).</Description>
                <ThumbnailUrl>thumb.jpg</ThumbnailUrl>
            </ImageSet>
        </ForegroundImageSet>
    </Place>
</Folder>"""

    @pytest.mark.skipif("not HAS_AVM")
    def test_avm(self):
        from xml.etree import ElementTree as etree

        # Hack for macOS: XML textualization is ever-so-slightly different.

        wtml = self.AVM_WTML

        if sys.platform == "darwin":
            wtml = wtml.replace('Dec="-42.58752472831171"', 'Dec="-42.587524728311706"')

        expected = etree.fromstring(wtml)

        cli.entrypoint(
            [
                "tile-study",
                "--avm",
                "--outdir",
                self.work_path(),
                test_path("geminiann11015a.jpg"),
            ]
        )

        with open(self.work_path("index_rel.wtml"), "rt", encoding="utf8") as f:
            observed = etree.fromstring(f.read())

        assert_xml_elements_equal(observed, expected)

    # Blah: this is the same as AVM_WTML, minus the metadata fields that aren't
    # preserved in the FITS export.
    FITS_WCS_WTML = """<?xml version='1.0' encoding='UTF-8'?>
<Folder Browseable="True" Group="Explorer" Name="Toasty" Searchable="True">
    <Place
        Angle="0"
        AngularSize="0"
        DataSetType="Sky"
        Dec="-42.58752472831171"
        Magnitude="0"
        Name="Toasty"
        Opacity="100"
        RA="23.269985215493794"
        Rotation="0"
        Thumbnail="thumb.jpg"
        ZoomLevel="0.3370990847923382"
    >
        <ForegroundImageSet>
            <ImageSet
                BandPass="Visible"
                BaseDegreesPerTile="0.07726507807936124"
                BaseTileLevel="0"
                BottomsUp="False"
                CenterX="349.049780614"
                CenterY="-42.5874939584"
                DataSetType="Sky"
                ElevationModel="False"
                FileType=".png"
                Generic="False"
                Name="Toasty"
                OffsetX="1.886354445296938e-05"
                OffsetY="2.4372703865669735e-05"
                Projection="Tan"
                QuadTreeMap=""
                Rotation="-138.99999999999943"
                Sparse="True"
                StockSet="False"
                TileLevels="3"
                Url="{1}/{3}/{3}_{2}.png"
                WidthFactor="2"
            >
                <ThumbnailUrl>thumb.jpg</ThumbnailUrl>
            </ImageSet>
        </ForegroundImageSet>
    </Place>
</Folder>"""

    def test_fits_wcs(self):
        from xml.etree import ElementTree as etree

        # Platform hack: XML textualization is ever-so-slightly different.

        wtml = self.FITS_WCS_WTML

        if sys.platform in ("darwin", "win32"):
            wtml = wtml.replace('Dec="-42.58752472831171"', 'Dec="-42.587524728311706"')

        expected = etree.fromstring(wtml)

        cli.entrypoint(
            [
                "tile-study",
                "--fits-wcs",
                test_path("geminiann11015a_wcs.fits"),
                "--outdir",
                self.work_path(),
                test_path("geminiann11015a.jpg"),
            ]
        )

        with open(self.work_path("index_rel.wtml"), "rt", encoding="utf8") as f:
            observed = etree.fromstring(f.read())

        assert_xml_elements_equal(observed, expected)
